//  
//  Copyright (C) 2009,2010,2011 Bartek thindil Jasicki
// 
//  This file is part of Grubng
// 
//  Grubng is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
// 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
// 

using System;
using System.Text;
using System.Text.RegularExpressions;

namespace urlsdb
{
	
	/// <summary>
	/// Class provide functions for manipulate URL's
	/// </summary>
	internal sealed class ParseURLs : IDisposable
	{
		/// <summary>
		/// Used to generate SHA256 hash of URL
		/// </summary>
		System.Security.Cryptography.SHA1Managed SHhash;
		/// <summary>
		/// SHA256 hash of URL
		/// </summary>
		StringBuilder key;
		/// <summary>
		/// Used to generate SHA256 hash of URL
		/// </summary>
		byte[] HashValue;
		/// <summary>
		/// Used to generate SHA256 hash of URL
		/// </summary>
		byte[] PureHash;
		/// <summary>
		/// If true, all disposable fields are disposed.
		/// </summary>
		bool disposed;
		
		/// <summary>
		/// Standard class constructor.
		/// </summary>
		public ParseURLs()
		{
			this.SHhash = new System.Security.Cryptography.SHA1Managed();
			this.key = new StringBuilder();
		}
		
		/// <summary>
		/// Function manipulate URL - check it correctness, drop to lower case host.
		/// </summary>
		/// <param name="url">
		/// A <see cref="System.String"/> URL to parse.
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/> parsed URL or String.Empty if URL is invalid.
		/// </returns>
		public static string ParseURL(string url)
		{
			//Remove http(s):// from URL
			int index = url.IndexOf("//");
			if (index > 0)
			{
				url = url.Remove(0, (index + 2));
			}
			//Check correctness of URL
			if (!Regex.IsMatch(url, @"^[a-z0-9]+([a-z0-9\=\-\.\?\,\'\/\\\+&amp;%\$#_~]*)?$", RegexOptions.IgnoreCase))
			{
				return String.Empty;
			}
			//Conver to lower case host name
			index = url.IndexOf('/');
			string host, path;
			if (index > 0)
			{
				host = url.Substring(0, index);
				path = url.Substring(index);
			}
			else
			{
				host = url;
				path = "/";
			}
			host = host.ToLower();
			//Validate hostname
			if (!Regex.IsMatch(host, @"^[a-z0-9]+([a-z0-9\-\.]*)?$", RegexOptions.IgnoreCase))
			{
				return String.Empty;
			}
			//Remove PHPSESID from path
			if (path.Length > 1)
			{
				path = Regex.Replace(path, @"(&amp;|\?|&)*PHPSESSID=\w+", String.Empty, RegexOptions.IgnoreCase);
			}
			url = host + path;
			url = System.Web.HttpUtility.UrlEncode(host + path);
			url = url.Replace("%2f", "/");
			url = url.Replace("%2c", ",");
			url = url.Replace("%3d", "=");
			url = url.Replace("%3f", "?");
			url = url.Replace("%26", "&");
			url = url.Replace("%3b", ";");
			url = url.Replace("%3a", ":");
			url = url.Replace("%40", "@");
			url = url.Replace("%23", "#");
			url = url.Replace("%7e", "~");
			url = url.Replace("%252f", "/");
			return url;
		}
		
		/// <summary>
		/// Function return SHA1 hash of URL
		/// </summary>
		/// <param name="url">
		/// A <see cref="System.String"/> URL to hash
		/// </param>
		/// <returns>
		/// A <see cref="System.String"/> SHA1 hash of URL
		/// </returns>
		public string GetHash(string url)
		{
			if (this.disposed)
			{
				throw new ObjectDisposedException(GetType().Name);
			}
			if (this.key.Length > 0)
			{
				this.key.Remove(0, key.Length);
			}
			this.PureHash = Encoding.UTF8.GetBytes(url);
			this.HashValue = SHhash.ComputeHash(this.PureHash);
			foreach(byte b in this.HashValue) 
			{
				this.key.Append(String.Format("{0:x2}", b));
			}
			return this.key.ToString();
		}
		
		/// <summary>
		/// Function dispose unmanaged resources.
		/// </summary>
		public void Dispose()
		{
			this.Dispose(true);
		}
		
		/// <summary>
		/// Function dispose unmanaged resources. Private function.
		/// </summary>
		/// <param name="disposing">
		/// A <see cref="System.Boolean"/> if true, dispose all unmanaged resources.
		/// </param>
		void Dispose(bool disposing)
		{
			if (!this.disposed)
			{
				this.SHhash = null;
				this.HashValue = null;
				this.PureHash = null;
				this.disposed = true;
				disposing = true;
			}
		}
	}
}
